
open Ast
open Typing
open Closure
open Mips

(* schma de compilation :
   - $s0 sauvegarde le $ra de main ; on n'y touche pas
   - $t0 temporaire pour les calculs
   - $a0 = argument de fonction, $a1 = adresse de la clture
*)

let globals = Hashtbl.create 17

(* on renomme les symboles pour viter les clash avec les instructions MIPS *)
let symb x = "S__" ^ x

(* labels pour les branchements *)
let new_label =
  let r = ref 0 in
  fun () -> incr r; "Label_" ^ string_of_int !r

let strings = Hashtbl.create 17
let alloc_string =
  let r = ref 0 in
  fun s ->
    incr r;
    let x = "String__" ^ string_of_int !r in
    Hashtbl.add strings x s;
    x

(* destination for the value of an expression *)
type dest = Stack | Register of register

let pushr r = sub sp sp oi 4 ++ sw r areg (0, sp)
let popr r = lw r areg (0, sp) ++ add sp sp oi 4
let pop1 = add sp sp oi 4

let constant dest n = match dest with
  | Stack -> li t0 n ++ pushr t0
  | Register r -> li r n

module Smap = Map.Make(String)

type env = {
  fpmax : int ref;     (* taille du cadre en octets *)
  fp : int;            (* prochain emplacement libre / $fp, en octets *)
  locals : int Smap.t; (* variable locale -> dplacement / $fp, en octets *)
  clvars : int Smap.t; (* variable cloture -> dplacement / $a1, en octets *)
}

let empty_env () =
  { fpmax = ref 0; fp = 0; locals = Smap.empty; clvars = Smap.empty }

(* destructure la valeur contenue dans le registre r selon un pattern p *)
let rec dest_pattern ?env r p = match p.tpatt_desc with
  | TP_any ->
      nop
  | TP_ident x -> begin match env with
      | None (* var globale *) ->
          sw r alab (symb x)
      | Some env (* var locale *) ->
          assert (Smap.mem x env.locals);
          let ofs = Smap.find x env.locals in
          sw r areg (-ofs, fp)
    end
  | TP_tuple pl ->
      let i = ref (-4) in
      let element acc p =
        i := !i + 4;
        pushr r ++
        lw r areg (!i, r) ++ dest_pattern ?env r p ++
        popr r ++ acc
      in
      List.fold_left element nop pl

let store r = function
  | Register r' when r' = r -> nop
  | Register r' -> move r' r
  | Stack -> pushr r

(* alloue une variable locale sur la pile *)
let alloc_local env x =
  let fp = env.fp in
  let fp' = fp + 4 in
  if fp' > !(env.fpmax) then env.fpmax := !(env.fpmax) + 4;
  { env with fpmax = env.fpmax; fp = fp'; locals = Smap.add x fp env.locals }

(* positionne les variables de clture *)
let alloc_clvars env cenv =
  let i = ref 0 in
  { env with clvars =
    List.fold_left (fun m x -> i := !i + 4; Smap.add x !i m) env.clvars cenv }

let rec alloc_pattern env p = match p.tpatt_desc with
  | TP_any -> env
  | TP_ident x -> alloc_local env x
  | TP_tuple pl -> List.fold_left alloc_pattern env pl

let binop op r1 r2 rdst = match op with
  | Badd | Bsub | Bmul | Bdiv ->
      let a = match op with
        | Badd -> add | Bsub -> sub | Bmul -> mul | Bdiv -> div
        | Bor|Band|Bge|Bgt|Ble|Blt|Bneq|Beq -> assert false
      in
      a rdst r1 oreg r2
  | Beq | Bneq | Blt | Ble | Bgt | Bge ->
      let s = match op with
        | Beq -> seq | Bneq -> sne
        | Blt -> slt | Ble -> sle | Bgt -> sgt | Bge -> sge
        | Bor|Band|Bdiv|Bmul|Bsub|Badd -> assert false
      in
      s rdst r1 oreg r2
  | Band | Bor ->
      assert false

let access_local_var env x dest =
  try
    let ofs = Smap.find x env.locals in
    begin match dest with
      | Stack -> lw t0 areg (-ofs, fp) ++ pushr t0
      | Register r -> lw r areg (-ofs, fp)
    end
  with Not_found ->
    let ofs = Smap.find x env.clvars in
    begin match dest with
      | Stack ->
          lw t0 areg (4, fp) ++ lw t0 areg (ofs, t0) ++ pushr t0
      | Register r -> lw t0 areg (4, fp) ++ lw r areg (ofs, t0)
    end

let alloc_closure cenv =
  let size = 4 * (1 + List.length cenv) in
  li a0 size ++ li v0 9 ++ syscall

let fill_closure env f cenv =
  let m = (alloc_clvars env cenv).clvars in
  let store_local acc x =
    access_local_var env x (Register t0) ++
    (let ofs = Smap.find x m in sw t0 areg (ofs, v0)) ++ acc
  in
  List.fold_left store_local nop cenv ++
  la t0 alab (symb f) ++ sw t0 areg (0, v0)

let rec expr env dest e = match e.cexpr_desc with
  | Ccte (Cunit | Cbool false) | Cnil -> constant dest 0
  | Ccte (Cbool true) -> constant dest 1
  | Ccte (Cint n) -> constant dest n
  | Ccte (Cstring s) ->
      let x = alloc_string s in
      (match dest with
        | Stack -> la t0 alab x ++ pushr t0
        | Register r -> la r alab x)
  | Capp (e1, e2) ->
      expr env Stack e2 ++
      expr env (Register a1) e1 ++ popr a0 ++
      lw t0 areg (0, a1) ++ jalr t0 ++ store v0 dest
  | Cglobal x -> begin match dest with
      | Stack -> lw t0 alab (symb x) ++ pushr t0
      | Register r -> lw r alab (symb x)
    end
  | Clocal x ->
      access_local_var env x dest
  | Cunop (Unot, e) ->
      expr env (Register t0) e ++
      seq t0 t0 oi 0 ++ store t0 dest
  | Cunop (Uminus, e) ->
      expr env (Register t0) e ++ neg t0 t0 ++ store t0 dest
  | Cif (e1, e2, e3) ->
      let l_else = new_label () in
      let l_done = new_label () in
      expr env (Register t0) e1 ++
      beqz t0 l_else ++
      expr env dest e2 ++
      b l_done ++ label l_else ++
      expr env dest e3 ++
      label l_done
  | Cbinop (Band, e1, e2) ->
      let lab = new_label () in
      expr env (Register t0) e1 ++
      beqz t0 lab ++
      expr env (Register t0) e2 ++
      label lab ++
      store t0 dest
  | Cbinop (Bor, e1, e2) ->
      let lab = new_label () in
      expr env (Register t0) e1 ++
      bnez t0 lab ++
      expr env (Register t0) e2 ++
      label lab ++
      store t0 dest
  | Cbinop (op, e1, e2) ->
      expr env Stack e1 ++ expr env (Register t0) e2 ++ popr t1 ++
        (match dest with
          | Register r -> binop op t1 t0 r
          | Stack -> binop op t1 t0 t0 ++ pushr t0)
  | Cclos (f, cenv) ->
      alloc_closure cenv ++
      fill_closure env f cenv ++ store v0 dest
  | Clet (false, p, e1, e2) ->
      let env' = alloc_pattern env p in
      let code1 = expr env (Register t0) e1 in
      code1 ++ dest_pattern ~env:env' t0 p ++ expr env' dest e2
  | Clet (true, p, {cexpr_desc=Cclos (f, cenv)}, e2) ->
      let env = alloc_pattern env p in
      alloc_closure cenv ++
      dest_pattern ~env v0 p ++ fill_closure env f cenv ++ expr env dest e2
  | Clet (true, _, _, _) ->
      assert false
  | Ctuple el ->
      let i = ref (-4) in
      let element e acc =
        i := !i + 4;
        expr env (Register t0) e ++ popr v0 ++
        sw t0 areg (!i, v0) ++ pushr v0 ++ acc
      in
      let size = 4 * List.length el in
      li a0 size ++ li v0 9 ++ syscall ++
      pushr v0 ++ (* sauvegarde sur la pile *)
      List.fold_right element el nop ++
      (match dest with Stack -> nop | Register r -> popr r)
  | Ccons (e1, e2) ->
      li a0 8 ++ li v0 9 ++ syscall ++
      pushr v0 ++
      expr env (Register t0) e1 ++ popr v0 ++ sw t0 areg (0, v0) ++
      pushr v0 ++
      expr env (Register t0) e2 ++ popr v0 ++ sw t0 areg (4, v0) ++
      store v0 dest
  | Cmatch (e1, e2, (ph, pt, e3)) ->
      let l_cons = new_label () in
      let l_done = new_label () in
      expr env (Register t0) e1 ++
      bnez t0 l_cons ++
      (* nil *)
      expr env dest e2 ++ b l_done ++
      (* cons *)
      label l_cons ++
      let env = alloc_pattern (alloc_pattern env ph) pt in
      lw t1 areg (0, t0) ++ dest_pattern ~env t1 ph ++
      lw t1 areg (4, t0) ++ dest_pattern ~env t1 pt ++
      expr env dest e3 ++
      label l_done

let fundef (f, cenv, p, e) =
  let env = empty_env () in
  let env = alloc_clvars env cenv in
  let env = alloc_pattern env p in
  let code = expr env (Register v0) e in
  let frame_size = !(env.fpmax) + 12 in
    label (symb f) ++
    sub sp sp oi frame_size ++ (* alloue la frame *)
    sw fp areg (frame_size - 4, sp) ++   (* sauve $fp *)
    sw ra areg (frame_size - 8, sp) ++   (* sauve $ra *)
    add fp sp oi (frame_size - 16) ++ (* $fp = ... *)
    sw a1 areg (4, fp)                (* sauve $a1 = la cloture *)
  ++
  dest_pattern ~env a0 p ++
  code ++
    lw ra areg (8, fp) ++   (* restaure $ra *)
    lw fp areg (12, fp) ++   (* restaure $fp *)
    add sp sp oi frame_size ++ (* dsalloue la frame *)
    jr ra

let letdef env (is_rec, p, e) =
  let rec alloc_globals p = match p.tpatt_desc with
    | TP_any -> ()
    | TP_ident x -> Hashtbl.replace globals x ()
    | TP_tuple pl -> List.iter alloc_globals pl
  in
  alloc_globals p;
  expr env (Register v0) e ++ dest_pattern v0 p

let compile_list compile l =
  List.fold_right (fun x acc -> compile x ++ acc) l nop

(* cltures pour les primitives *)
let prims =
  Hashtbl.fold
    (fun x _ acc ->
      acc ++ la t0 alab ("Code_" ^ x) ++ sw t0 alab ("Closure_" ^ x)
      ++ la t0 alab ("Closure_" ^ x) ++ sw t0 alab (symb x))
    primitives nop

let file (dl,fl) =
  let env = empty_env () in
  let decls = compile_list (letdef env) dl in
  let funs = compile_list fundef fl in
  let frame_size = !(env.fpmax) in
  { text =
      label "main" ++
      move s0 ra ++ (* sauvegarde de $ra *)
      sub sp sp oi frame_size ++ (* alloue la frame *)
      add fp sp oi (frame_size - 4) ++ (* $fp = ... *)
      prims ++
      decls ++
      add sp sp oi frame_size ++ (* dsalloue la frame *)
      move ra s0 ++ (* restaure $ra *)
      jr ra ++
      funs ++
      (* primitives *)
        label "Code_print_int" ++
        li v0 1 ++
        syscall ++
        jr ra ++
        label "Code_print_string" ++
        li v0 4 ++
        syscall ++
        jr ra ++
        label "Code_print_newline" ++
        li v0 4 ++
        la a0 alab "newline" ++
        syscall ++
        jr ra ++
        label "Code_read_int" ++
        li v0 5 ++
        syscall ++
        jr ra ++
        label "Code_read_line" ++
        la a0 alab "Buffer_read_line" ++
        li a1 100 ++ (* limite de taille pour read_line *)
        li v0 8 ++
        syscall ++
        jr ra
      ;
    data =
      Hashtbl.fold
        (fun p _ acc -> label (symb p) ++ dword [0] ++
          label ("Closure_" ^ p) ++ dword [0] ++ acc)
        primitives nop
      ++
      label "newline" ++ asciiz "\n" ++
      label "Buffer_read_line" ++ space 32
      ++
      Hashtbl.fold (fun x _ l -> label (symb x) ++ dword [0] ++ l) globals nop
      ++
      Hashtbl.fold (fun x s l -> label x ++ asciiz s ++ l) strings nop
  }

